/* Pavel Surynek, 2008 */
/* pavel.surynek@mff.cuni.cz */
/* tbox.cpp */

/*
  An implementation of a new algorithm called theta-BOX for solving
  sliding box puzzles.
 */

#include "slider.h"
#include "tbox.h"
#include "obox.h"


void solution2fine_solution(SLIDER *slider, const SOLUTION &solution, FINE_SOLUTION &fine_solution)
{
	for (int i = 0; i < solution.m_M; ++i)
	{
		FINE_SOLUTION phase_solution;

		switch (solution.m_moves[i])
		{
		case CYCLE_1_LEFT:
		{
			cycle_1_rotate_L(slider, phase_solution);
			break;
		}
		case CYCLE_1_RIGHT:
		{
			cycle_1_rotate_R(slider, phase_solution);
			break;
		}
		case CYCLE_2_LEFT:
		{
			cycle_2_rotate_L(slider, phase_solution);
			break;
		}
		case CYCLE_2_RIGHT:
		{
			cycle_2_rotate_R(slider, phase_solution);
			break;
		}
		case CYCLE_3_LEFT:
		{
			cycle_3_rotate_L(slider, phase_solution);
			break;
		}
		case CYCLE_3_RIGHT:
		{
			cycle_3_rotate_R(slider, phase_solution);
			break;
		}
		default:
		{
			assert(false);
			break;
		}
		}
		fine_solution.insert(fine_solution.end(), phase_solution.begin(), phase_solution.end());
	}
}


bool solve_slider_tbox(SLIDER *slider, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool &in_database, double timeout)
{
	SLIDER copy;

	copy_slider(slider, &copy);

	int p;

	in_database = true;

	for (int i = 1; i < copy.m_V; ++i)
	{
		if (copy.m_vrtxs[i] != copy.m_goal[i])
		{
			for (int j = 1; j < copy.m_V; j++)
			{
				if (copy.m_vrtxs[j] == copy.m_goal[i])
				{
					p = j;
					break;
				}
			}
			RECORDS_map::iterator record = database.m_records.find(RECORD_KEY(slider->m_A, slider->m_B, slider->m_C, i, p));

			if (record != database.m_records.end())
			{
				execute_fine_solution(&copy, record->second);
				solution.insert(solution.end(), record->second.begin(), record->second.end());
			}
			else
			{
				if (timeout <= 0.0)
				{
					return false;
				}

				clock_t begin, end;

				SLIDER deep_copy;
				copy_slider(&copy, &deep_copy);

				printf("%d <---> %d\n", i, p);
				for (int k = 0; k < deep_copy.m_V; ++k)
				{
					deep_copy.m_goal[k] = deep_copy.m_vrtxs[k];
				}
				swap_goal(i, p, &deep_copy);

				SOLUTION coarse_solution;
				FINE_SOLUTION fine_solution;
				init_solution(1024, &coarse_solution);

				begin = clock();
				bool solved = solve_slider_ID_PL(&deep_copy, &coarse_solution, timeout);
				
				end = clock();
				double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;
				timeout -= elapsed;
				
				if (solved)
				{
					in_database = false;
					printf("***\n");

					solution2fine_solution(&copy, coarse_solution, fine_solution);
					database.m_records[RECORD_KEY(slider->m_A, slider->m_B, slider->m_C, i, p)] = fine_solution;
					solution.insert(solution.end(), fine_solution.begin(), fine_solution.end());
				}
				else
				{
					destroy_slider(&copy);
					destroy_slider(&deep_copy);
					return false;
				}
				
				destroy_solution(&coarse_solution);
				
				if (timeout < 0.0)
				{
					destroy_slider(&copy);
					destroy_slider(&deep_copy);
					return false;
				}

				destroy_slider(&deep_copy);
			}
		}
	}
	destroy_slider(&copy);

	return true;
}


void generate_solution_database(const char *filename)
{
	SLIDER slider;
	SOLUTION_DATABASE database;

	int id;
	double time_out = 10.0;

	int c1_b, c2_a, c2_b, r1, r2;

	char filename_no[1024];
	int no = 0;

	sprintf(filename_no, "%s_%d.txt", filename, no);
	database.load(filename_no);

	while (1)
	{
		printf("Computing for timeout:%f...\n", time_out);

		id = 1;

		for (c1_b = 3; c1_b < 64; ++c1_b)
		{
			for (c2_a = 1; c2_a <= c1_b; ++c2_a)
			{
				for (c2_b = c1_b + 1; c1_b - c2_a + c2_b - c1_b + 1 <= c1_b; ++c2_b)
				{
					for (r1 = 1; r1 < c2_b; ++r1)
					{
						for (r2 = r1 + 1; r2 < c2_b + 1; ++r2)
						{

							init_slider(0, c1_b, c2_a, c2_b, &slider);

//							printf("abc:%d,%d,%d\n", slider.m_A, slider.m_B, slider.m_C);
//							print_slider(&slider);

							int oddness = slider.m_C1 % 2 || slider.m_C2 % 2 || slider.m_C3 % 2;
							
							clock_t begin, end;
							
							if (oddness && (c1_b != 5 || c2_a != 3 || c2_b != 6))
							{
								printf("%d_%d_%d\n", c1_b, c2_a, c2_b);
								printf("%d,%d\n", r1, r2);

								swap_goal(r1, r2, &slider);
								STEPS = 0;

								begin = clock();

								FINE_SOLUTION solution;

								bool in_data;
								solve_slider_tbox(&slider, database, solution, in_data, time_out);

								end = clock();
								double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;

								sprintf(filename_no, "%s_%d.txt", filename, no);

								if (!in_data)
								{
									database.save(filename_no);
								}
								
								no = (no + 1) % 10;
							}

							destroy_slider(&slider);
							++id;
						}
					}
				}
			}
		}

		time_out *= 2;
	}
}


bool solve_slider_3_tbox(SLIDER *slider, int X, int Y, int Z, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool &in_database, double timeout)
{
	if (timeout <= 0.0)
	{
		return false;
	}

	SLIDER copy;
	copy_slider(slider, &copy);
	
	RECORDS_3_map::iterator record = database.m_3_records.find(RECORD_3KEY(slider->m_A, slider->m_B, slider->m_C, X, Y, Z));

	if (record != database.m_3_records.end())
	{
		in_database = true;

		execute_fine_solution(&copy, record->second);
		solution.insert(solution.end(), record->second.begin(), record->second.end());
	}
	else
	{
		if (timeout <= 0.0)
		{
			return false;
		}

		clock_t begin, end;
		
		SLIDER deep_copy;
		copy_slider(&copy, &deep_copy);
		
		printf("%d ---> %d ---> %d\n", X, Y, Z);
		for (int k = 0; k < deep_copy.m_V; ++k)
		{
			deep_copy.m_goal[k] = deep_copy.m_vrtxs[k];
		}
		swap_goal(X, Y, &deep_copy);
		swap_goal(X, Z, &deep_copy);
		
		SOLUTION coarse_solution;
		FINE_SOLUTION fine_solution;
		init_solution(1024, &coarse_solution);
		
		begin = clock();
		bool solved = solve_slider_ID_PL(&deep_copy, &coarse_solution, timeout);
		
		end = clock();
		double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;
		timeout -= elapsed;
		
		if (solved)
		{
			in_database = false;
			printf("***\n");

			solution2fine_solution(&copy, coarse_solution, fine_solution);
			database.m_3_records[RECORD_3KEY(slider->m_A, slider->m_B, slider->m_C, X, Y, Z)] = fine_solution;
			solution.insert(solution.end(), fine_solution.begin(), fine_solution.end());
		}
		else
		{
			destroy_slider(&copy);
			destroy_slider(&deep_copy);
			return false;
		}
		
		destroy_solution(&coarse_solution);
		
		if (timeout < 0.0)
		{
			destroy_slider(&copy);
			destroy_slider(&deep_copy);
			return false;
		}
		destroy_slider(&deep_copy);
	}

	destroy_slider(&copy);

	return true;
}


void generate_solution_3_database(const char *filename)
{
	SLIDER slider;
	SOLUTION_DATABASE database;

	int id;
	double time_out = 10.0;
	char filename_no[1024];

	int c1_b, c2_a, c2_b, r1, r2, r3;
	int no = 0;

	sprintf(filename_no, "%s_%d.txt", filename, no);
	database.load(filename_no);

	while (1)
	{
		printf("Computing for timeout:%f...\n", time_out);

		id = 1;

		for (c1_b = 3; c1_b < 64; ++c1_b)
		{
			for (c2_a = 1; c2_a <= c1_b; ++c2_a)
			{
				for (c2_b = c1_b + 1; c1_b - c2_a + c2_b - c1_b + 1 <= c1_b; ++c2_b)
				{
					for (r1 = 1; r1 < c2_b + 1; ++r1)
					{
						for (r2 = 1; r2 < c2_b + 1; ++r2)
						{
							if (r2 != r1)
							{
								for (r3 = 1; r3 < c2_b + 1; ++r3)
								{
									if (r3 != r1 && r3 != r2)
									{
										init_slider(0, c1_b, c2_a, c2_b, &slider);
							
										clock_t begin, end;
							
										if (c1_b != 5 || c2_a != 3 || c2_b != 6)
										{
											printf("%d_%d_%d\n", c1_b, c2_a, c2_b);
											printf("%d,%d,%d\n", r1, r2, r3);

											swap_goal(r1, r2, &slider);
											swap_goal(r1, r3, &slider);
											STEPS = 0;

											begin = clock();
											
											FINE_SOLUTION solution;
											bool in_data;

											solve_slider_3_tbox(&slider, r1, r2, r3, database, solution, in_data, time_out);
											end = clock();
											double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;
											sprintf(filename_no, "%s_%d.txt", filename, no);

											if (!in_data)
											{
												database.save(filename_no);
											}

											no = (no + 1) % 10;
										}
										destroy_slider(&slider);
										++id;
									}
								}
							}
						}
					}
				}
				
			}
		}
		time_out *= 2;
	}
}


void generate_solution_database_Revised(const char *filename)
{
	SLIDER slider;
	SOLUTION_DATABASE database;

	int id;
	double time_out = 2.0;

	int r1, r2, A, B, C;

	char filename_no[1024];
	int no = 0;

	sprintf(filename_no, "%s_%d.txt", filename, no);
	database.load(filename_no);

	int limit = 6;

	while (1)
	{
		printf("Computing for timeout:%f...\n", time_out);

		id = 1;

		for (A = 1; A < limit; ++A)
		{
			for (B = 2; B < limit; ++B)
			{
				for (C = 1; C < limit; ++C)
				{
					int total = A + B + C;

					for (r1 = 1; r1 < total - 1; ++r1)
					{
						for (r2 = r1 + 1; r2 < total; ++r2)
						{
							init_slider(A, B, C, &slider);

//							printf("abc:%d,%d,%d\n", slider.m_A, slider.m_B, slider.m_C);
//							print_slider(&slider);
//							int oddness = slider.m_C1 % 2 || slider.m_C2 % 2 || slider.m_C3 % 2;
							
							clock_t begin, end;
							
							printf("ABC:%d,%d,%d\n", A, B, C);
							printf("%d,%d\n", r1, r2);

							swap_goal(r1, r2, &slider);
							STEPS = 0;

							begin = clock();

							FINE_SOLUTION solution;

							bool in_data;
							solve_slider_tbox(&slider, database, solution, in_data, time_out);
							end = clock();
							double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;

							sprintf(filename_no, "%s_%d.txt", filename, no);

							if (!in_data)
							{
								database.save(filename_no);
							}
							no = (no + 1) % 10;

							destroy_slider(&slider);
							++id;
						}
					}
				}
			}
		}

		time_out *= 2;
	}
}


void generate_solution_3_database_Revised(const char *filename)
{
	SLIDER slider;
	SOLUTION_DATABASE database;

	int id;
	double time_out = 2.0;
	char filename_no[1024];

	int r1, r2, r3, A, B, C;
	int no = 0;

	sprintf(filename_no, "%s_%d.txt", filename, no);
	database.load(filename_no);

	int limit = 8;

	while (1)
	{
		printf("Computing for timeout:%f...\n", time_out);

		id = 1;

		for (A = 1; A < limit; ++A)
		{
			for (B = 2; B < limit; ++B)
			{
				for (C = 1; C < limit; ++C)
				{
					int total = A + B + C;

					for (r1 = 1; r1 < total; ++r1)
					{
						for (r2 = 1; r2 < total; ++r2)
						{
							if (r2 != r1)
							{
								for (r3 = 1; r3 < total; ++r3)
								{
									if (r3 != r1 && r3 != r2)
									{
										init_slider(A, B, C, &slider);
							
										clock_t begin, end;
										
										printf("ABC:%d,%d,%d\n", A, B, C);
										printf("%d,%d,%d\n", r1, r2, r3);

										swap_goal(r1, r2, &slider);
										swap_goal(r1, r3, &slider);
										STEPS = 0;
											
										begin = clock();
											
										FINE_SOLUTION solution;
										bool in_data;
										solve_slider_3_tbox(&slider, r1, r2, r3, database, solution, in_data, time_out);
										end = clock();
										double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;
										sprintf(filename_no, "%s_%d.txt", filename, no);

										if (!in_data)
										{
											database.save(filename_no);
										}
										no = (no + 1) % 10;
										
										destroy_slider(&slider);
										++id;
									}
								}
							}
						}
					}
				}
				
			}
		}
		time_out *= 2;
	}
}
