/* Pavel Surynek, 2008 */
/* pavel.surynek@mff.cuni.cz */
/* tbox.h */

/*
  An implementation of a new algorithm called theta-BOX for solving
  sliding box puzzles.
 */

#ifndef __TBOX_H__
#define __TBOX_H__

#include <map>

#include "slider.h"
#include "mit.h"


class RECORD_KEY
{
public:
	RECORD_KEY()
	{
		m_A = m_B = m_C = 0;
		m_X = 1;
		m_Y = 2;
	}

	RECORD_KEY(int A, int B, int C, int X = 1, int Y = 2)
	{
		m_A = A;
		m_B = B;
		m_C = C;
		m_X = X;
		m_Y = Y;
	}

	RECORD_KEY(const RECORD_KEY &key)
	{
		m_A = key.m_A;
		m_B = key.m_B;
		m_C = key.m_C;
		m_X = key.m_X;
		m_Y = key.m_Y;
	}

	const RECORD_KEY& operator=(const RECORD_KEY &key)
	{
		m_A = key.m_A;
		m_B = key.m_B;
		m_C = key.m_C;
		m_X = key.m_X;
		m_Y = key.m_Y;

		return *this;
	}

	bool operator<(const RECORD_KEY &key) const
	{
		return ((m_A < key.m_A) ? true : ((m_A > key.m_A) ? false :
			((m_B < key.m_B) ? true : ((m_B > key.m_B) ? false :
			 ((m_C < key.m_C) ? true : ((m_C > key.m_C) ? false :
			  ((m_X < key.m_X) ? true : ((m_X > key.m_X) ? false :
			   ((m_Y < key.m_Y) ? true : ((m_Y > key.m_Y) ? false : false))))))))));
	}

	void save(FILE *f) const
	{
		fprintf(f, "ABC:%d,%d,%d\n", m_A, m_B, m_C);
		fprintf(f, "XY:%d,%d\n", m_X, m_Y);
	}

	void load(FILE *f)
	{
		fscanf(f, "ABC:%d,%d,%d\n", &m_A, &m_B, &m_C);
		fscanf(f, "XY:%d,%d\n", &m_X, &m_Y);
	}

public:
	int m_A, m_B, m_C;
	int m_X, m_Y;
};


class RECORD_3KEY
{
public:
	RECORD_3KEY()
	{
		m_A = m_B = m_C = 0;
		m_X = 1;
		m_Y = 2;
		m_Z = 3;
	}

	RECORD_3KEY(int A, int B, int C, int X = 1, int Y = 2, int Z = 3)
	{
		m_A = A;
		m_B = B;
		m_C = C;
		m_X = X;
		m_Y = Y;
		m_Z = Z;
	}

	RECORD_3KEY(const RECORD_3KEY &key)
	{
		m_A = key.m_A;
		m_B = key.m_B;
		m_C = key.m_C;
		m_X = key.m_X;
		m_Y = key.m_Y;
		m_Z = key.m_Z;
	}

	const RECORD_3KEY& operator=(const RECORD_3KEY &key)
	{
		m_A = key.m_A;
		m_B = key.m_B;
		m_C = key.m_C;
		m_X = key.m_X;
		m_Y = key.m_Y;
		m_Z = key.m_Z;

		return *this;
	}

	bool operator<(const RECORD_3KEY &key) const
	{
		return ((m_A < key.m_A) ? true : ((m_A > key.m_A) ? false :
			((m_B < key.m_B) ? true : ((m_B > key.m_B) ? false :
			 ((m_C < key.m_C) ? true : ((m_C > key.m_C) ? false :
			  ((m_X < key.m_X) ? true : ((m_X > key.m_X) ? false :
			   ((m_Y < key.m_Y) ? true : ((m_Y > key.m_Y) ? false :
			    ((m_Z < key.m_Z) ? true : ((m_Z > key.m_Z) ? false : false))))))))))));
	}

	void save(FILE *f) const
	{
		fprintf(f, "ABC:%d,%d,%d\n", m_A, m_B, m_C);
		fprintf(f, "XYZ:%d,%d,%d\n", m_X, m_Y, m_Z);
	}

	void load(FILE *f)
	{
		fscanf(f, "ABC:%d,%d,%d\n", &m_A, &m_B, &m_C);
		fscanf(f, "XYZ:%d,%d,%d\n", &m_X, &m_Y, &m_Z);
	}

public:
	int m_A, m_B, m_C;
	int m_X, m_Y, m_Z;
};


typedef std::map<RECORD_KEY, FINE_SOLUTION, std::less<RECORD_KEY> > RECORDS_map;
typedef std::map<RECORD_3KEY, FINE_SOLUTION, std::less<RECORD_3KEY> > RECORDS_3_map;


class SOLUTION_DATABASE
{
public:
	SOLUTION_DATABASE()
	{
	}

	SOLUTION_DATABASE(const SOLUTION_DATABASE &database)
	{
		m_records = database.m_records;
	}

	const SOLUTION_DATABASE& operator=(const SOLUTION_DATABASE &database)
	{
		m_records = database.m_records;

		return *this;
	}

	void save(FILE *f) const
	{
		fprintf(f, "Solution database\n");
		fprintf(f, "Size:%d\n", m_records.size());

		for (RECORDS_map::const_iterator rec = m_records.begin(); rec != m_records.end(); ++rec)
		{
			rec->first.save(f);
			save_fine_solution(f, rec->second);
		}

		fprintf(f, "3 Size:%d\n", m_3_records.size());

		for (RECORDS_3_map::const_iterator rec3 = m_3_records.begin(); rec3 != m_3_records.end(); ++rec3)
		{
			rec3->first.save(f);
			save_fine_solution(f, rec3->second);
		}
	}

	void load(FILE *f)
	{
		int size;

		fscanf(f, "Solution database\n");
		fscanf(f, "Size:%d\n", &size);

		for (int i = 0; i < size; ++i)
		{
			RECORD_KEY key;
			key.load(f);
			FINE_SOLUTION fine_sol;
			load_fine_solution(f, fine_sol);

			m_records[key] = fine_sol;
		}

		fscanf(f, "3 Size:%d\n", &size);

		for (int i = 0; i < size; ++i)
		{
			RECORD_3KEY key3;
			key3.load(f);
			FINE_SOLUTION fine_sol;
			load_fine_solution(f, fine_sol);

			m_3_records[key3] = fine_sol;
		}
	}

	int save(const char *filename) const
	{
		FILE *f;

		f = fopen(filename, "w");

		if (f == NULL)
		{
			printf("Cannot open file %s for writing.\n", filename);
			return -1;
		}

		save(f);
		fclose(f);

		return 0;
	}

	int load(const char *filename)
	{
		FILE *f;

		f = fopen(filename, "r");

		if (f == NULL)
		{
			printf("Cannot open file %s for reading.\n", filename);
			return -1;
		}

		load(f);
		fclose(f);

		return 0;
	}

public:
	RECORDS_map m_records;
	RECORDS_3_map m_3_records;
};


void solution2fine_solution(SLIDER *slider, const SOLUTION &solution, FINE_SOLUTION &fine_solution);
bool solve_slider_tbox(SLIDER *slider, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool &in_database, double timeout = 10.0);
bool solve_slider_3_tbox(SLIDER *slider, int X, int Y, int Z, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool &in_database, double timeout = 10.0);

void generate_solution_database(const char *filename);
void generate_solution_3_database(const char *filename);

void generate_solution_database_Revised(const char *filename);
void generate_solution_3_database_Revised(const char *filename);

#endif
